        /********************************************************
         *                                                      *
         *      ARTIFICIAL   NEURAL   NETWORK   SOFTWARE        *
         *                                                      *
         *      By Andrea Hadjiprocopis                         *
         *                                                      *
         *              1992/93                                 *
         *                                                      *
         ********************************************************/



#include "XMain.h"
#include "TMain.h"

void    ParChanOutInt( Process *p, Channel *c, int data );
void    ParChanMesIn( Process *p, Channel *c );
double	f_inv( double y );

void 	ParChanOut( Process *p, Channel *c, double data, Channel *safe )
{	
	p = p;

if( CRESET )	ChanReset(c);
if( DEBUG )	printf("\nSTARTING CHANOUT IN PARALLEL CHANEL= %p\n", c); 

	ChanOutChanFail( c, (void *) &data, sizeof( double ), safe);

if( DEBUG )	printf("DONE CHANOUT, CHANEL= %p, DATA= %lf\n",c,data);
}
/* 	******************************************************************************    */
void    ParChanOutInt( Process *p, Channel *c, int data )
{
        p = p;

if( CRESET )        ChanReset(c);
if( DEBUG ) printf("\nSTARTING CHANOUT INT   PARALLEL CHANEL= %p\n", c);

        ChanOutInt( c, data);

if( DEBUG ) printf("DONE CHANOUT INT,  CHANEL= %p, DATA= %d\n",c,data);

}
/*      ******************************************************************************    */

void 	FeedInputLayer( int attrib )
{
	/* This is a process to distribute the inputs to all the next layer nodes, to BROADCAST	*
	   It can be done either by ParChanOut, directly, the Training data, or by using the
	   function FeedInputLayer 
	   Here it is done by selecting RandomLY training examples and passing them to the nextlayer in parallel*/

	Neuron 	*n;  	/* A list of all input neurons			*/
	Process **p;	/* A 1D list to the processes to run chanout in parallel, a 1D list to store seed out procs*/
	double	the_output;  
	int i, j, k, iterations = 0, exemp = -1, ex_num;

	if( (n=(Neuron *)calloc(CFTable[0], sizeof(Neuron ) )) == NULL )
                ErrorHandler( 13, MALLOC_ERROR, 0 );
	if( (p=(Process **)calloc( CFTable[0]*CFTable[1]+1, sizeof(Process *) )) == NULL )
                ErrorHandler( 13, MALLOC_ERROR, 1 );

	if( attrib == EXAMINATION ){
		max_iters = 1;
		ex_num = 1;
	} else ex_num = num_exemplars;
	for(i=0;i<CFTable[0];i++){
		if( (n[i]=NCP[ GetRelativeOffset(0, i) ]) == NULL ) 
			ErrorHandler( 13, NEXT_NODE_ERROR, i );
	}
	if( DEBUG )        printf("IN INPUT num_exemplars = %d\nmax_layers = %d\ninpwid = %d\ninphei = %d\noutwid = %d\nouthei = %d\n",
                			num_exemplars, max_layers, inpwid, inphei, outwid, outhei);
	if( DEBUG )        for(i=0;i<max_layers;i++) printf("CFTable[%d] = %d\n", i, CFTable[i]);

/* Keep sending input to next layer until reach maximum # of epochs */
while( iterations++ < max_iters ){
	exemp = -1;
	while( ++exemp < ex_num ){
		ProcWait(500);
		for(i=0;i<CFTable[0];i++){
			if( attrib == LEARNING ) the_output = Training[ exemp ].Input[ i ];
			else the_output = examination[i]; /* for every INPUT NODE */
			if( DEBUG )	printf("AT ITE=%d. FIL:(%d) a-vector=%d, training=%lf\n",iterations, i,exemp, the_output);
			for(j=0;j<(n[i])->outnum;j++){
				ProcWait(500);
				if( CRESET ) ChanReset( (n[i]->out)[j] );
				if( (p[i*((n[i])->outnum) + j]=ProcAlloc( ParChanOut, 1024, 4, (n[i]->out)[j], the_output, ((n[i])->FP_safe_receive)[j],NULL )) == NULL )
					ErrorHandler( 13, ALLOCATE_PROC, i );
			} /* for every HIDDEN NODE CONNECTED TO Ith INPUT NODE */
		}
		p[ i*j ] = NULL; /* list terminated by NULL */	

		ProcParList( p );  
	
		for(k=0;k<i*j;k++)	ProcAllocClean( p[k] );

		ProcWait(5000);
		if( exemp%1 == 0 ) printf(" +++EXEMPLAR SO FAR: %d\n", exemp);
	} /* While exemp: ALL EXEMPLARS FINISHED. GO TO NEXT ITERARTION */
	if( (iterations%1==0) && !(DEBUG) ) printf(" --- ITERATIONS SO FAR: %d\n", iterations);
} /* While iterations: ALL ITERARTIONS FININSHED. GAME OVER*/

if( !DEBUG && (attrib==LEARNING) ) 	printf(" --- ITERATIONS SO FAR: %d\n", iterations-1);/* The last one*/

}
 



void    HiddenLayerProcessing( Neuron n, int layer, int node,  int attrib )
{
/* This is the core process of a Neuron, depending on the attribute, attrib == FORW_PROP,
   the process will take all inputs, sum them apply threshold and transfer function
   and then release them to the output ( actually to all the output channels ).
   If the attrib == BACK_PROP then it will look for the errors coming from the 
   output channels and propagate them back to the inputs
*/

	Process **p,	 /* a pointer to the list of processes to be run in parallel in order to chanout in parallel*/
		**p_error = NULL;
	int 	*status;/* each bit in the status word is set to 1 if the
                        corresponding input channel has sent its value
                        it is zero from the begining. if a bit is set
                        already the relevant channel shouldn't be allowed
                        to give input again */ 
	int	*status_error = NULL;
        double	sum = 0.0, an_input = 0.0, deltasum = 0.0, delta = 0.0, to_send;
        int 	chanel = 0, Break = FALSE, i, count = 0, index = 0, iterations = 0, exemp = -1, ex_num;

	if( attrib == EXAMINATION ){
		max_iters = 1;
		ex_num = 1;
	} else ex_num = num_exemplars;
	if( (layer == INPUT) || (layer == max_layers-1) || (node<0) || (node>=CFTable[layer]) )
		ErrorHandler(8, WRONG_PARAMETERS, 0);

	if( (p=(Process **)calloc( 1+n->outnum, sizeof(Process *) )) == NULL )
		ErrorHandler( 8, MALLOC_ERROR, 2 );
	if( layer > 1 )
		if( (p_error=(Process **)calloc( 1+n->inpnum, sizeof(Process *) )) == NULL )
			ErrorHandler( 8, MALLOC_ERROR, 3 );

	if( (status=(int *)calloc( n->inpnum, sizeof(int) )) == NULL )
		 ErrorHandler( 8, MALLOC_ERROR, 4 );
	if( (status_error=(int *)calloc( n->outnum, sizeof(int) )) == NULL )
		 ErrorHandler( 8, MALLOC_ERROR, 5 );
	if( DEBUG )	printf("HIDDEN LAYER NODE(%d,%d)= %p,  CH1= %p, CH2=%p,CH3=%p\n",layer,node,n,n->inp[0], n->inp[1], n->inp[2]);	

while( iterations++ < max_iters ){
	exemp = -1;
	while( ++exemp < ex_num ){
		Break = FALSE;index = 0;chanel = -100;count = 0;sum = 0.0;deltasum = 0.0, delta = 0.0;
		for(i=0;i<n->inpnum;i++) {
			status[i] = 0;
			if( CRESET ) ChanReset( n->inp[i] );
		} /* reset status and channel each time u do a cycle */
		while( !Break ){
		/* Get the input from any input channel, finish when all input channels have sent ONCE ( Break = TRUE ) the weighted sum is being calculated once an input value is received */
			do {
				chanel = ProcSkipAltList( n->inp );
				if( index++ > PROCALT_TIMED_OUT ){
					if( ErrorHandler(8, ACTION_TIMED_OUT, PROCALT_TIMED_OUT) ) index = 0; /* eh will return TRUE if user wants to do another loop of 4000 */
					else 	 Quit(0);	/* if not it wont come here because it will exit but never mind */
				}
			} while( (chanel<0) || (chanel>=n->inpnum) );
			if( (chanel<0) || (chanel>=n->inpnum) )
				ErrorHandler(8, EMPTY_CHAN_LIST, chanel);
			if( status[chanel] == 0) {
				if( DEBUG ) printf("HID LAYER: I WILL HAVE DATA SOON COMING OUT FROM CHAN=%p OF NODE(%d,%d)=%p\n",(n->inp)[ chanel ],layer,node,n);
				ChanIn( (n->inp)[ chanel ], &an_input, sizeof(double));
				ChanOutInt( (n->FP_safe_send)[chanel], 1);
				status[chanel] = 1;
				(n->input)[ chanel ] = an_input; /* Store the input from chanel to the array so it can be used later on BACK_P pass */
				if( DEBUG ) printf("HID LAYER: GOT THAT DATA FROM CHAN=%p OF NODE(%d,%d)=%p. DATA= %lf\n",(n->inp)[ chanel ], layer,node,n,an_input);
				sum += an_input * (n->weight[ chanel ]);
       				count++;
			}
			if( count >= n->inpnum ) Break = TRUE;
		} /* WHILE NOT Break */
		/* Exit from the receive loop, by now all input channels have sent ONCE, weighted sum in sum, input values in n->input */
		
		(n->output) = f(sum, n->threshold);/* Now the output is ready to be sent, sent in PARALLEL */
		for(i=0;i<(n->outnum);i++){
			if( CRESET ) ChanReset( n->out[i] );
			if( (p[i]=ProcAlloc( ParChanOut, 1024, 4, n->out[i], n->output, (n->FP_safe_receive)[i], NULL )) == NULL )
				ErrorHandler(8, ALLOCATE_PROC, i);
			if( DEBUG ) printf("SENDING TO CH(%d)= %p DATA=%lf\n",i,(n->out)[i], n->output);
		}
		p[i] = NULL; /* list should be terminated by NULL */

		/* Send output in parallel here everything stops until everything is sent */
		ProcParList( p ); /* ChanOut in parallel */
		for(i=0;i<n->outnum;i++) {
			ProcAllocClean( p[i] );
			p[i] = NULL;
		} /* for loop */

		/* Now start the BACK_P pass: for NON-OUTPUT layers ONLY ( output layer for output nodes next function ) */
		if( attrib == LEARNING ){
			Break = FALSE;index = 0;chanel = -10000;count = 0;deltasum = 0.0;
			for(i=0;i<n->outnum;i++) {
				status_error[i] = 0;
				if( CRESET )	ChanReset( n->errI[i] );
			}
			while( !Break ){
				do {
					chanel = ProcSkipAltList( n->errI );
					if( index++ > PROCALT_TIMED_OUT ){
						if( ErrorHandler(8, ACTION_TIMED_OUT, PROCALT_TIMED_OUT) ) index = 0;
						else 	 Quit(0);
					}
				} while( (chanel<0) || (chanel>=n->outnum) );
				if( (chanel<0) || (chanel>=n->outnum) ){
					printf("CHANEL PROB AT HID NODE=%p\n",n);
					ErrorHandler(8, EMPTY_CHAN_LIST, chanel);
				}
				if( status_error[chanel] == 0) {
					if( DEBUG ) printf("HID LAYER: BACK PROPAGATION PHASE: I WILL HAVE DATA SOON COMING OUT FROM CHAN=%p OF NODE(%d,%d)=%p\n",(n->errI)[ chanel ],layer,node,n);
					ChanIn( (n->errI)[ chanel ], &an_input, sizeof(double));
					ChanOutInt( (n->BP_safe_send)[chanel], 1);
					status_error[chanel] = 1;
					if( DEBUG ) printf("HID LAYER: BACK PROPAGATION PHASE: GOT THAT DATA FROM CHAN=%p OF NODE(%d,%d)=%p. DATA= %lf\n",(n->errI)[ chanel ], layer,node,n,an_input);
					deltasum += an_input;
       					count++;
				}
				if( count >= n->outnum ) Break = TRUE;
			} /* WHILE NOT Break */
			/* Exit from the receive loop, by now all input channels have sent ONCE, weighted sum in sum, input values in n->input */
			(n->delta) = deltasum * (n->output) * (1.0-n->output);
			/* the output of the current node times the derivative of the activation f */

			/* Send our delta*weight to previous node before we update our weights */
			if( layer > 1 ){ /* First hidden layer will not send any error back , BP is finished */
				for(i=0;i<n->inpnum;i++){
					if( CRESET ) ChanReset( (n->errO)[i] );
					to_send = (n->delta) * (n->weight)[i];
					if( (p_error[i]=ProcAlloc( ParChanOut, 1024, 4, (n->errO)[i], to_send, (n->BP_safe_receive)[i],NULL )) == NULL )
						ErrorHandler(8, ALLOCATE_PROC, i);
				}
				p_error[i] = NULL;
				ProcParList( p_error );
				for(i=0;i<n->inpnum;i++){
					ProcAllocClean( p_error[i] );
					p_error[i] = NULL;
				}
			}
			for(i=0;i<n->inpnum;i++)
				(n->weight)[i] = Lamda * (n->weight)[i] + Beta * (n->delta) * ( layer == 1   ?   (n->input)[i] : f_inv(n->input[i]) );
		} /* IF LEARNING */
	} /* WHILE exemplars loop */
} /* WHILE iterations Loop */
	
}			


void 	OutputLayerProcessing( Neuron n, int node, double *target_data, int target_num, int attrib )
{

	Process	**p_error;
	int 	*status; 
			/* each bit in the status word is set to 1 if the
                        corresponding input channel has sent its value
                        it is zero from the begining. if a bit is set
                        already the relevant channel shouldn't be allowed
                        to give input again */ 
        double	sum = 0.0, an_input = 0.0, to_send;
        int 	chanel = 1000000, Break = FALSE, i, count = 0, index = 0, iterations = 0, exemp = 0, ex_num;

	if( (node<0) || (node>=CFTable[max_layers-1]) ) 
		ErrorHandler( 12, OUT_OF_RANGE, node );
        if( (status=(int *)calloc( n->inpnum, sizeof(int) )) == NULL )
                ErrorHandler( 12, MALLOC_ERROR, 1 );
	if( (p_error=(Process **)calloc( 1+n->inpnum, sizeof(Process *) )) == NULL )
		ErrorHandler( 12, MALLOC_ERROR, 2 );

	if( attrib == EXAMINATION ){
		max_iters = 1;
		ex_num = 1;
	} else ex_num = num_exemplars;
	if( DEBUG )	printf("OUTPUT LAYER NODE(%d)= %p,INPCHANS= %p, %p, %p, CHANOUT=%p\n",node,n, n->inp[0], n->inp[1], n->inp[2], n->out[0]);	

while( iterations++ < max_iters ){
	exemp = -1;
	while( ++exemp < ex_num ){		
		Break = FALSE;count = 0;sum = 0.0;
	        for(i=0;i<n->inpnum;i++) {
        	        status[i] = 0;
			if( CRESET )	ChanReset( n->inp[i] );
        	} /* reset status and channel each iter */	
		while( !Break ){
			index = 0;
			do {
				chanel = ProcSkipAltList( n->inp );
				if( index++ > PROCALT_TIMED_OUT ){
					if( ErrorHandler(12, ACTION_TIMED_OUT,PROCALT_TIMED_OUT) ) index = 0; /* eh will return TRUE if user wants to do another loop of 4000 */
					else 	 Quit(0);	/* if not it wont come here because it will exit but never mind */
				}
			} while( (chanel < 0) || (chanel>=n->inpnum) ); /* do */
			if( (chanel < 0) || (chanel > MAX_NODE) )
				ErrorHandler(12, EMPTY_CHAN_LIST,0);

			if( DEBUG ) printf("OUT LAYER NODE(%d)=%p INPUT CHANEL RESPONDS AT CH= %p, chanel=%d, STATUS=%d\n",node,n,(n->inp)[ chanel ], chanel,status[chanel] );
			if( status[chanel] == 0) {
				if( DEBUG ) printf("OUT LAYER: I WILL HAVE DATA SOON COMING OUT FROM CHAN=%p OF NODE(%d)=%p\n",(n->inp)[ chanel ],node,n);
				ChanIn( (n->inp)[ chanel ], &an_input, sizeof(double));
				ChanOutInt( (n->FP_safe_send)[chanel], 1);
				status[chanel] = 1;
				(n->input)[chanel] = an_input;
				if( DEBUG ) printf("OUT LAYER: GOT THAT DATA FROM CHAN=%p OF NODE(%d)=%p. DATA= %lf\n",(n->inp)[ chanel ],node,n,an_input);
				sum += an_input * (n->weight[ chanel ]);
       				count++;
			} /* status */
			if( count >= n->inpnum ) Break = TRUE;
		} /* WHILE NOT Break */
		(n->output) = f(sum, n->threshold);
		if( attrib == LEARNING ){
			if( DEBUG )	printf("OL(%d), exemp=%d, TARG=%lf\n", node,exemp, Training[ exemp ].Target[ node ]);
			(n->delta) = (Training[ exemp ].Target[ node ] - n->output) * (n->output) * (1.0 - n->output);
			/* Send our delta*weight to previous node before we update our weights */
			for(i=0;i<n->inpnum;i++){
			if( CRESET )	ChanReset( (n->errO)[i] );
				to_send = (n->delta) * (n->weight)[i];
				ProcWait(400);
				if( (p_error[i]=ProcAlloc( ParChanOut, 2048, 4, (n->errO)[i], to_send, (n->BP_safe_receive)[i], NULL)) == NULL )
					ErrorHandler(8, ALLOCATE_PROC, i);
			}
			p_error[i] = NULL;
			ProcParList( p_error );
			for(i=0;i<n->inpnum;i++){
				ProcAllocClean( p_error[i] );
				p_error[i] = NULL;
				(n->weight)[i] = Lamda * (n->weight)[i] + Beta * (n->delta) * f_inv(n->input[i]);
			}
		} /* IF LEARNING */
	
		/* now send the output,not the error, to the error server */
		if( CRESET )		ChanReset(n->out[0]);
		ProcWait(200);
		ChanOut(n->out[0], (void *) &(n->output), sizeof(double) );
		if( DEBUG )	printf("SEND OUTPUT = %lf, at EXE = %d and ITE = %d\n", n->output, exemp, iterations );

	} /* while exemp loop */ 

} /* WHILE iterations loop */
	

	ProcWait(1000);
	ChanOutInt( message, 1);      
}


double  f( double sum, double threshold )
{
/* this is the transfer function of the neuron node, usually is a sigmoid but can take
   some other types as well like tanh or sign or...
*/
        return( (double )(1/(1 + exp( (double )(-sum -threshold) ))) );
}


double f_inv( double  y )

{
	return( (double )log( (double )(y / (1.0-y)) ) );
}
