        /********************************************************
         *                                                      *
         *      ARTIFICIAL   NEURAL   NETWORK   SOFTWARE        *
         *                                                      *
         *      By Andrea Hadjiprocopis                         *
         *                                                      *
         *              1992/93                                 *
         *                                                      *
         ********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <channel.h>
#include <process.h>
#include "ErrorHandler.h"

/* 	DEFINES		*/
#define TERM_NOT        (void *)0xFFF   /* termination number for alloc procs, chans */
#define INPUT		0
#define	LEARNING	1
#define EXAMINATION	2
#define MAX_WEIGHT	1.0 /* The maximum weight value that can be output from the random gen machine */
#define MAX_THRESH	1.1 /* The same for threshold values */

#define TRUE		1
#define	FALSE		0

#define	DISCRETE	'D'
#define	CONTINUOUS	'C'

#define	STACK_SIZE	50000

#define MAX_TIME        12000
#define RETRY_TIME      8000
#define RETRY_BASE      1000
#define MAX_NODE        100
#define PROCALT_TIMED_OUT 4000000

#define	INPUT_NODES_WIDTH	1
#define INPUT_NODES_HEIGTH	4
#define OUTPUT_NODES_WIDTH	1
#define OUTPUT_NODES_HEIGHT	2
#define INPUT_NODES		4
#define OUTPUT_NODES		2
#define	HIDDEN_NODES		4
#define	NUM_EXEMPLARS		2
#define	MAX_LAYERS		3
#define	BETA			1.4
#define	LAMDA			1.0


/* MACROS AND VARIOUS TRICKS */
#define randomize(a) (srand( (unsigned )time(NULL) + a )) /* applies the time( up to second ) as seed to rand */
#define ABS(a)  (a < 0  ? -a:a) /* absolute value coz the buildin abs is useless */
#define	ROUND_UP(a)	(double )( (int )(a+0.5) )



/* 	EXTERNAL VARIABLE AND TYPEDEFS	*/
extern struct 		NEURON {
	int 	layer, node;	/* The coordinates of the neuron 					*/
        double  *weight;        /* A pointer to a list of weights connected to the INPUT  of the neuron */
                                /* The number of weights is by definition the number of OUTPUT channels */
        double  threshold;      /* The value of the threshold associated to the specific neuron 	*/

	double  *input;		/* A pointer to the list of all the input values to neuron		*/

	int     outnum;         /* The number of output coonections from the node , ie # of *out        */
        Channel **out;          /* A pointer to the list of the output channels                         */
        int     inpnum;         /* The number of output coonections from the node , ie # of *inp        */
        Channel **inp;          /* A pointer to the list of the input channels                          */
	Channel **errI;		/* A pointer to the list of the channels receiving the delta from next L*/
	Channel **errO;		/* A pointer to the list of the channels sending the delta to previous L*/

	Channel **seed;		/* A list to the channels transmitting the seed only for FI and OL(single)*/
	double  output;		/* the value of the last output computed*/
	double	delta;		/* the error relevant to the last output computed above */

	Channel	**FP_safe_receive, /*receive acknowledge from next layer next*/
		**FP_safe_send,	/* send acknowledge back to previous layer size previous*/
		**BP_safe_receive,
		**BP_safe_send,

		*abort;		
		};
  /* by this any neuron can be implemented, also an input is implemented as weight = 1, inp = NULL, inpnum=0*/
  /* an output can be implemented as out=NULL, outnum=NULL		*/
  /* an input is considered as a unit which actually produces the inputs( so no need of input chan )	*/
  /* an output is considered as a unit which actually manipulates the outputs so no need of output channel*/
  /* FOR THE TIME BEING THAT IS*/
typedef struct NEURON   *Neuron;        /* A pointer to the structure NEURON for return reasons */

struct 		TRAINING {
	double	*Target;
	double	*Input; 
		};

extern  struct 	TRAINING   	*Training, 
                		*AnalogTempTrainingData,
                		*DigitalTempTrainingData;


extern 	Channel *message;
extern 	int 	DEBUG, CRESET,
		max_layers, 	max_layers_OLD, 	/* the maximum number of layers shoulb be ddefined in MAIN	*/
		*CFTable,	*CFTable_OLD,   	/* says the max number of nodes in each layer*/
		num_exemplars, 	num_exemplars_OLD,	/* the number of training vectors */
		max_iters,				/* the number of iterations */
		*RandomVector;
extern	double  *examination,
		*single_output,
		Beta, Lamda;				/* the rate of learning as defined in the hebbian rule */
		
extern 	Neuron 	*NCP;       /* A one dimensional array of pointers each pointing to a structure NEURON containing data for any single neuron
			       that has been created and registered. the index to NCP is only one, therefore to implement the 2Dim array structure
			       (layer,node) we use the function GetRelativeOffset(layer,node). In this way we economize memory space since not all
			       the layers will have the same number of nodes.etc */
/* 	FUNCTION DECLARATIONS	*/

/* FLAGS */
extern	int     First_time_setup_flag,
		CFT_has_changed_flag;
/*	allocs.c	*/
extern 	int 	AllocateChans( /*va_alist*/ );

/* 	ehandler.c	*/
extern 	int     ErrorHandler( unsigned function_code, int description, unsigned useful_number );
extern 	void 	Quit( int exit_code );
extern 	void 	FreeNeuronData( Neuron n );
extern	void	DisplayError(char *str);

/* 	neuron.c	*/
extern 	Neuron  CreateNeuron( int layer, int node, int input_number, int output_number );
extern	void    RegisterNeuron( int layer, int node, Neuron n );
extern	int     GetRelativeOffset( int layer, int node );

/* 	calculations.c 	*/
extern 	void    HiddenLayerProcessing( Neuron n, int layer, int node, int attrib );
extern  double  f( double sum, double threshold );
extern	void    ParChanOut( Process *p, Channel *c, double data, Channel *safe );
extern	void    OutputLayerProcessing( Neuron n, int node, double *target_data, int target_num, int attrib );
extern	void    FeedInputLayer( int attrib );

/* 	main.c		*/
extern  double 	Random(  double max );
extern	int	RandomInt( int max_limit );
extern	void 	prSER(Process *p, int attribute, char type);
extern 	void	prIN( Process *p, int attrib );
extern 	void 	prHL( Process *p, int layer, int node, int attrib );
extern	void 	prOUT( Process *p, int node, double *target, int vector_num, int attrib );
extern	void 	TransputerInitialize( void );

extern	void	CreateTheNeuralNetwork( void );

/* error server */
extern	void	ErrorServer(int attribute, char type );
