        /********************************************************
         *                                                      *
         *      ARTIFICIAL   NEURAL   NETWORK   SOFTWARE        *
         *                                                      *
         *      By Andrea Hadjiprocopis                         *
         *                                                      *
         *              1992/93                                 *
         *                                                      *
         ********************************************************/

#include "XMain.h"
#include "TMain.h"


Neuron 	CreateNeuron( int layer, int node, int input_number, int output_number )
{
/****************************************************************************************
 * This function creates a neuron which has input_number inputs, output_number outputs	*
 * its position is (layer, node) and has output_number weights each in every INPUT 	*
 * channel. It mallocs a structure of type NEURON allocates the channels and returns 	*
 * a pointer to that structure. the pointer may be used to define the connections	*
 * Procedure:										*
 *	1. Check if layer==INPUT or layer==OUTPUT.If yes NULL appropriately		*
 *	2. otherwise check for previous nodes output channels and make them our inputs  *
 * 	   of course this will take the first ao the second or.. of ecah previous node  *
 * 	3.  
 ****************************************************************************************/

	Neuron a_neuron;
	int i;

	if( (layer > max_layers) | (node > CFTable[ layer ]) ) 
		ErrorHandler( 6, OUT_OF_RANGE, 0); /* Out of specified map dimensions */

	if( (a_neuron=(Neuron )calloc(1,sizeof(struct NEURON))) == NULL ) 
		ErrorHandler( 6, MALLOC_ERROR, 8 );

	a_neuron->inpnum = input_number;
	a_neuron->outnum = output_number;
	a_neuron->delta  = 0.0;
	a_neuron->output = 0.0;
	a_neuron->layer  = layer;
	a_neuron->node   = node;
	a_neuron->threshold = Random(MAX_THRESH);


	if( (a_neuron->inp=(Channel **)calloc(input_number+1, sizeof(Channel *))) == NULL )
		ErrorHandler( 6, MALLOC_ERROR, 0 ); /* the input channels */
	if( (a_neuron->weight=(double *)calloc(input_number, sizeof(double))) == NULL )
		ErrorHandler( 6, MALLOC_ERROR, 1 ); /* store the weights  */
	if( (a_neuron->out=(Channel **)calloc(output_number+1, sizeof(Channel *))) == NULL )
		ErrorHandler( 6, MALLOC_ERROR, 2 ); /* the output channels*/

	if( layer == 0 ) {
		if( node == 0 ) {
                        if( ((a_neuron->seed)=(Channel **)calloc(CFTable[max_layers-1]+1,sizeof(Channel *))) == NULL )
				ErrorHandler( 6, MALLOC_ERROR, 3 ); /* the control*/
			for(i=0;i<CFTable[max_layers-1];i++)
				if( ((a_neuron->seed[i])=ChanAlloc()) == NULL ) /* Control Chanels for input node zero */
			       		ErrorHandler( 6, ALLOCATE_CHANS, 0 );
		}
		if( ((a_neuron->inp[0])=ChanAlloc()) == NULL ) /* Just ONE input chan for input layer */
                        ErrorHandler( 6, CREATE_NEURON_INP, node );
		if( CRESET ) ChanReset( a_neuron->inp[0] );
		a_neuron->inp[1] = NULL; /* list always must be terminated by NULL */
		a_neuron->inpnum = 1;  /* Just ONE input chan for input layer */
		a_neuron->weight[0] = 1.0; /*and no weights, ie gain is one */
	
	} else {  /* NOT AN INPUT LAYER */

		if( (a_neuron->input=(double *)calloc(input_number, sizeof(double))) == NULL )
			ErrorHandler( 6, MALLOC_ERROR, 4 ); /* store input values */
       		if( (a_neuron->errO=(Channel **)calloc(input_number+1, sizeof(Channel *))) == NULL )
               		ErrorHandler( 6, MALLOC_ERROR, 5 ); /* the deltas channel OUT ( to previous NODEs )*/
       		if( (a_neuron->errI=(Channel **)calloc(output_number+1, sizeof(Channel *))) == NULL )
               		ErrorHandler( 6, MALLOC_ERROR, 6 ); /* the deltas channel */
		/* ALL NODES EXCEPT INPUT HAVE errI. OUTPUT NODE USES ONLY errI[0] for the error server */
		for(i=0;i<output_number;i++){
			if( ((a_neuron->errI)[i]=ChanAlloc()) == NULL )
                               	ErrorHandler( 6, ALLOCATE_CHANS, i );
			if( CRESET ) ChanReset( (a_neuron->errI)[i] );
		}

		(a_neuron->errI)[i] = NULL; /* list always must be terminated by NULL */	

		if( layer == max_layers-1 ){ /* OUTPUT Layer */
       
			if( ((a_neuron->seed)=(Channel **)calloc(1,sizeof(Channel *))) == NULL )
				ErrorHandler( 6, MALLOC_ERROR, 7 ); /* the input channels */
			if( ((a_neuron->seed[0])=NCP[GetRelativeOffset(0,0)]->seed[node]) == NULL )
				ErrorHandler( 6, ALLOCATE_CHANS, i );
		}


		for(i=0;i<input_number;i++){ /* to find out how many inputs the node should have */
			if( ((a_neuron->inp)[i]=( (NCP[ GetRelativeOffset(layer-1,i) ] )->out)[ node ] ) == NULL )
				ErrorHandler( 6, PREVIOUS_NODE_ERROR, i );
			if( ((a_neuron->errO)[i]=( (NCP[ GetRelativeOffset(layer-1,i) ] )->errI)[ node ] ) == NULL )
        	        	ErrorHandler( 6, PREVIOUS_NODE_ERROR, i ); /* errO exists always in hidden, output nodes */
			if( CRESET ) ChanReset( (a_neuron->inp)[i] );
			if( CRESET ) ChanReset( (a_neuron->errO)[i] );
			(a_neuron->weight)[i] = Random(MAX_WEIGHT);
		}
		(a_neuron->inp)[i] = NULL; /* list always must be terminated by NULL */
		(a_neuron->errO)[i] = NULL;

	} /* else */

	for(i=0;i<output_number;i++) {
		if( ((a_neuron->out)[i]=ChanAlloc()) == NULL )
			ErrorHandler( 6, CREATE_NEURON_OUT, i );
		if( CRESET ) ChanReset( (a_neuron->out)[i] );
	}
	(a_neuron->out)[i] = NULL; /* list always must be terminated by NULL */
	

	/* INPUT LAYER : FP_receive, BP_send.
	   HIDDEN LAYER: FP_receive, FP_send, BP_receive, BP_send.
	   OUTPUT LAYER: FP_send, BP_receive. */

	if( layer != max_layers-1 ){ /* IF THIS IS NOT AN OUTPUT LAYER */

		if( (a_neuron->FP_safe_receive=(Channel **)calloc(output_number+1, sizeof(Channel *))) == NULL )
			ErrorHandler( 6, MALLOC_ERROR, 4 ); /* the input channels */
		for(i=0;i<output_number;i++)
			if( ((a_neuron->FP_safe_receive)[i]=ChanAlloc()) == NULL )
		       		ErrorHandler( 6, ALLOCATE_CHANS, 100+i );

		if( layer != 0 ){ 
			if( (a_neuron->FP_safe_send=(Channel **)calloc(input_number+1, sizeof(Channel *))) == NULL )
				ErrorHandler( 6, MALLOC_ERROR, 7 ); /* the input channels */
			for(i=0;i<input_number;i++)
				if( ((a_neuron->FP_safe_send)[i]=(NCP[ GetRelativeOffset(layer-1, i) ]->FP_safe_receive)[node]) == NULL )
			       		ErrorHandler( 6, ALLOCATE_CHANS, 200+i );
		}

	} else { /* IF THIS IS AN OUTPUT LAYER */
		if( (a_neuron->FP_safe_send=(Channel **)calloc(input_number+1, sizeof(Channel *))) == NULL )
			ErrorHandler( 6, MALLOC_ERROR, 8 ); /* the input channels */
		for(i=0;i<input_number;i++)
			if( ((a_neuron->FP_safe_send[i])=(NCP[GetRelativeOffset(layer-1,i)]->FP_safe_receive)[node]) == NULL )
			       		ErrorHandler( 6, ALLOCATE_CHANS, 300+i );
	}

	if( layer != 0 ){ /* INPUT LAYERS ARE NOT IN BP ALGORITHM */
		if( layer != 1 ){ /* THIS IS NOT A FIRST HIDDEN LAYER */
			if( (a_neuron->BP_safe_receive=(Channel **)calloc(input_number+1, sizeof(Channel *))) == NULL )
        	                ErrorHandler( 6, MALLOC_ERROR, 9 ); /* the input channels */
	                for(i=0;i<input_number;i++)
                	        if( ((a_neuron->BP_safe_receive)[i]=(NCP[GetRelativeOffset(layer-1,i)]->BP_safe_send)[node]) == NULL )
                        	        ErrorHandler( 6, ALLOCATE_CHANS, 400+i );
			if( layer != max_layers-1 ){ /* THIS IS NOT AN OUTPUT LAYER */
		                if( (a_neuron->BP_safe_send=(Channel **)calloc(output_number+1, sizeof(Channel *))) == NULL )
					ErrorHandler( 6, MALLOC_ERROR, 8 ); /* the input channels */
                		for(i=0;i<output_number;i++)
                        		if( ((a_neuron->BP_safe_send[i])=ChanAlloc()) == NULL ) 
	                       	ErrorHandler( 6, ALLOCATE_CHANS, 500+i );
			}
		} else { /* THIS IS A FIRST HIDDEN LAYER */
	                if( (a_neuron->BP_safe_send=(Channel **)calloc(output_number+1, sizeof(Channel *))) == NULL )
        	                ErrorHandler( 6, MALLOC_ERROR, 11 ); /* the input channels */
        	        for(i=0;i<output_number;i++)
				if( ((a_neuron->BP_safe_send[i])=ChanAlloc()) == NULL )
                                        ErrorHandler( 6, ALLOCATE_CHANS, 600+i );
		}
	}


	RegisterNeuron( layer, node, a_neuron );

if( DEBUG ){
        printf("\nNODE (%d,%d) has a pointer: %p\n",a_neuron->layer, a_neuron->node, a_neuron);
        printf("IN CHANS(%d): ", a_neuron->inpnum);
        for(i=0;i<a_neuron->inpnum;i++) printf(" %p, ",a_neuron->inp[i]);
        printf("\nOU CHANS(%d): ", a_neuron->outnum);
        for(i=0;i<a_neuron->outnum;i++) printf(" %p, ",a_neuron->out[i]);
        printf("\nWEIGHTS: ");
        for(i=0;i<a_neuron->inpnum;i++) printf(" %lf, ",a_neuron->weight[i]);
        printf("\nTHRESHOLD: %lf", a_neuron->threshold);
/*	printf("\nERROR INPUT CHANNEL: ");
	for(i=0;i<a_neuron->inpnum;i++) printf("  %p, ", a_neuron->errI[i]);
	printf("\nERROR OUTPUT CHANNEL: ");
	for(i=0;i<a_neuron->outnum;i++) printf("  %p, ", a_neuron->errO[i]);
 	printf("%lf, %lf\n", Random(MAX_THRESH), a_neuron->threshold);
*/
	printf("\nFORWARD RECEIVE CHANELS :");
	for(i=0;i<a_neuron->outnum;i++) printf("%p, ", a_neuron->FP_safe_receive[i]);
	printf("\nFORWARD SEND CHANELS :");
	for(i=0;i<a_neuron->inpnum;i++) printf("%p, ", a_neuron->FP_safe_send[i]);	

	printf("\nBACKPROP RECEIVE CHANELS :");
	for(i=0;i<a_neuron->inpnum;i++) printf("%p, ", a_neuron->BP_safe_receive[i]);
	printf("\nBACKPROP SEND CHANELS :");
	for(i=0;i<a_neuron->outnum;i++) printf("%p, ", a_neuron->BP_safe_send[i]);	

	puts("");
}
	return( a_neuron );
}


void 	RegisterNeuron( int layer, int node, Neuron n )
{

	(NCP[ GetRelativeOffset( layer, node ) ])  = n; /* store the pointer pointing to the information of that node */

/*	printf("REG(%d,%d)= %p, NCP=%d, F=%d\n",layer, node,NCP[ GetRelativeOffset( layer, node ) ], (NCP[ GetRelativeOffset( layer, node ) ])->inpnum, n->inpnum);
*/
} /* probably can be implemented as a macro? */


int	GetRelativeOffset( int layer, int node )
{
	int i, space = 0;


	if( layer != 0 ) for(i=0;i<=(layer-1);i++) space += CFTable[ i ];
	
	space += node;	/* now add the offfset of the node */
	
	return( space );
}



 

	


